/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.docware;

import java.util.Collection;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.regex.Pattern;

import casa.util.pdu.Attributes;

import casa.dodwan.util.DataFormatting;
import casa.dodwan.util.Keyable;
import casa.dodwan.util.StringGenerator;
import java.util.Properties;

/**
 * Descriptor of a DoDWAN message. A descriptor is composed of attributes, each
 * having a name and a value. A descriptor contains internal attributes (with
 * reserved names starting with an underscore character) and user-defined
 * attributes (with freely chosen names, but excluding reserved ones). All
 * attributes can be accessed with a generic getter/setter. 
 */
public class Descriptor {
    
    /**
     * @undocumented
     */
    Attributes attributes_ = null;
    
    
    
    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Set<Map.Entry<String, String>> getEntrySet() {
	return attributes_.entrySet();
    }

    // ------------------------------------------------------------
    /**
     * Returns the number of attributes in this <code>Descriptor</code>.
     * @return the number of attributes
     */
    public int getNbAttributes() {
	return attributes_.size();
    }

    // ------------------------------------------------------------
    /**
     * Returns the value of the attribute with the specified name in this
     * <code>Descriptor</code>, or <code>null</code> if this attribute does not
     * exist.
     * @param name the name of the attribute whose value should be returned
     * @return the value of this attribute (or <code>null</code> if this
     *         attribute does not exist)
     */
    public String getAttribute(String name) {
	return attributes_.get(name);
    }

    // ------------------------------------------------------------
    /**
     * Returns the names of the attributes defined in this
     * <code>Descriptor</code>.
     * @return the names of the attributes defined in this
     * <code>Descriptor</code>.
     */
    public Set<String> getNames() {
	return attributes_.keySet();
    }

    // ------------------------------------------------------------
    /**
     * Removes the attribute with the specified name from this
     * <code>Descriptor</code>.
     * @param name the name of the attribute that should be removed
     */
    public void removeAttribute(String name) {
	attributes_.remove(name);
    }

    // ------------------------------------------------------------
    /**
     * Sets a new attribute (or changes an existing one) in this
     * <code>Descriptor</code>..
     * @param name the name of the attribute
     * @param value the value assigned to this attribute
     */
    public void setAttribute(String name, String value) {
	attributes_.put(name, value);
    }

   // ------------------------------------------------------------
    /**
     * Creates an empty descriptor.
     */
    public Descriptor() {
	attributes_ = new Attributes();
    }
	
    
  // ------------------------------------------------------------
    /**
     * Creates a descriptor from a Attributes object.
     */
     public Descriptor(Attributes attrs) {
        attributes_ = attrs;
    }
    
   
    
   // ------------------------------------------------------------
     /**
     * Produces an Attribute object from the descriptor 
     * @return the Attribute object containing the list of attribute assignments
     */
    public Attributes asAttributes() {
        return (Attributes) attributes_;
    }

    
    
     // ------------------------------------------------------------
    /**
     * Creates a descriptor from a string containing a list of attributes.
     * The string should be composed of comma-separated list of attribute assignments 
     * of the form: key=value,key=value,...
     * @param sattrs the string containing attributes.
     * @return the descriptor formed from this list of attributes.
     */
    public static Descriptor fromAttributes(String sattrs) {
	Descriptor desc = new Descriptor();
	// Extracting descriptor attributes and values
	String[] descAttr = sattrs.split(",");
	for (int i = 0; i < descAttr.length; i++) {
	    String[] attr = descAttr[i].split("=");
	    desc.setAttribute(attr[0], attr[1]);
	}
	return desc;
    }
    
    // ------------------------------------------------------------
    /**
     * Produces a string composed of a comma-separated list of attribute assignments 
     * of the form: key=value,key=value,...
     * @return the string containing the list of attribute assignments
     */
    public String toAttributes() {
        return DataFormatting.toString(attributes_, ",");
    }

    
    
       
      
    // ------------------------------------------------------------
    /**
     * Creates a descriptor from a list of Java properties.
     * @param props the properties object containing attributes.
     * @return the descriptor formed from this list of attributes.
     */
    public static Descriptor fromProperties(Properties props) {
	Descriptor desc = new Descriptor();
        for (String name: props.stringPropertyNames()) {
            desc.setAttribute(name, props.getProperty(name));
        }
	return desc;
    }
    
    
    
    // ------------------------------------------------------------
    /**
     * Produces a Java properties object from the descriptor 
     * @return the properies object containing the same attributes
     */
    public Properties toProperties() {
        Properties props = new Properties();
        
        for (Object obj: attributes_.keySet()) {
	    String key = (String)obj;
	    props.setProperty(key, getAttribute(key));
	}
        return props;
    }
    
    
    
    // ------------------------------------------------------------
    /**
     * Creates a descriptor as a copy of the specified descriptor.
     * @param other the other descriptor that must be copied.
     */
    public Descriptor(Descriptor other) {

	attributes_ = new Attributes();
	for (Object obj: other.attributes_.keySet()) {
	    String key = (String)obj;
	    setAttribute(key, other.getAttribute(key));
	}
    }

    // ------------------------------------------------------------
    /**
     * Merges this descriptor with another one. The resulting descriptor
     * contains  all the attributes present in this descriptor, completed by the 
     * attributes of the other descriptor that are not present 
     * in this descriptor.
     * @param other the other descriptor 
     * @return The merged descriptor
     */
    public Descriptor mergeWith(Descriptor other) {
	
	for (Object obj: other.attributes_.keySet()) {
	    String key = (String)obj;
	    if (getAttribute(key) == null)
		setAttribute(key, other.getAttribute(key));
	}
	return this;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void setAttribute(String name, List<String> values) {

	setAttribute(name, DataFormatting.toString(values, ","));
    }
	
    // ------------------------------------------------------------
     /**
     * @undocumented
     */
    public void setAttribute(String name, Set<String> values) {

	setAttribute(name, DataFormatting.toString(values, ","));
    }

    // ------------------------------------------------------------
    /**
     * Returns the document identifier of this descriptor.
     * @return the value of the "_docid" attribute.
     *         or <code>null</code> if this attribute does not exist.
     */
    public String getDocumentId() {
	return getAttribute("_docid");
    }

    // ------------------------------------------------------------
    /**
     * Sets the document id (or changes the existing one) in this descriptor.
     * @param documentId the new value of the "_docid" attribute.
     */
    public void setDocumentId(String documentId) {
	setAttribute("_docid", documentId);
    }

    // ------------------------------------------------------------
    /**
     * Returns the descriptor's deadline, if any.
     * @return the deadline of this descriptor, as specified in the "_dline" attribute.
     *         or <code>null</code> if this attribute is not defined.
     */
    public Date getDeadline() {
        String dline = getAttribute("_dline");
        if (dline == null)
            return null;
        else
            return new Date(Long.parseLong(dline));
    }

    // ------------------------------------------------------------
    /**
     * Sets the descriptor's deadline.
     * @param deadline the deadline to be set as the "_dline" attribute.
     */
    public void setDeadline(Date deadline) {
        if (deadline == null)
            removeAttribute("_dline");
        else
            setAttribute("_dline", String.valueOf(deadline.getTime()));
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Descriptor getMatchingAttributes(Descriptor regex)
	throws Exception {

	Descriptor result = new Descriptor();
	
	for (Object obj: regex.attributes_.keySet()) {
	    String key = (String)obj;
	    String localValue = getAttribute(key);
	    if (localValue != null) {
		String regexValue = regex.getAttribute(key);
		if (Pattern.matches(regexValue, localValue))
		    result.setAttribute(key, localValue);
	    }
	}
	return ((result.getNbAttributes() > 0)?result:null);
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public Set<Descriptor> getMatchingDescriptors(Collection<Descriptor> descs) {

	Set<Descriptor> result = new HashSet<Descriptor>();

	for (Descriptor elt : descs) {
	    try {
		if (elt.matches(this))
		    result.add(elt);
	    }
	    catch (Exception e) {
	    }
	}
	return result;
    }

    // ------------------------------------------------------------
    /**
     * @undocumented
     */
   public Set<String> getMatchingKeys(Collection<Descriptor> descs) {

	Set<String> result = new HashSet<String>();

	//  		System.out.println("Testing pattern " + this + " against:");
	for (Descriptor elt : descs) {
	    try {
		boolean match = elt.matches(this);
		//  			System.out.println("  " + elt + " -> " + match);
		if (match)
		    result.add(elt.getDocumentId());
	    }
	    catch (Exception e) {
	    }
	}
	return result;
    }
    
    // ------------------------------------------------------------
    /**
     * Tests if this descriptor is equal to another descriptor. They are equal 
     * if all the attributes (name and value) they contains are equal.
     * @param other The other descriptor to compare with
     * @return True if this descriptor is equal to the other descriptor, false otherwise.
     */
    public boolean equals(Descriptor other) {

	if (attributes_.size() != other.attributes_.size())
	    return false;

	for (Object obj: attributes_.keySet()) {
	    String key = (String)obj;
	    if (! getAttribute(key).equals(other.getAttribute(key)))
		return false;
	}

	return true;
    }

    // ------------------------------------------------------------
    /**
     * Matches all the attributes in this descriptor against the
     * attributes contained in the specified descriptor.
     * <p>
     * Attribute matching is performed assuming that the attributes defined in
     * <code>regex</code> define Java regular expressions, against which the
     * attributes defined in this descriptor should be matched.
     * This descriptor is said to "match" <code>regex</code> if,
     * for any attribute defined in <code>regex</code>, there exists an
     * attribute with the same name in this descriptor whose value
     * matches that of the <code>regex</code>'s attribute.
     * @param regex the descriptor whose attributes define
     *        regular expressions that must be matched by the attributes of this
     *        descriptor.
     * @return <code>True</code> if all attributes in <code>regex</code> are
     *         matched by attributes in this descriptor, false
     *         otherwise.
     */
    public boolean matches(Descriptor regex) {

	for (Object obj: regex.attributes_.keySet()) {
	    String key = (String)obj;
	    String localValue = getAttribute(key);
	    if (localValue == null)
		return false;
	    String regexValue = regex.getAttribute(key);
	    if (! Pattern.matches(regexValue, localValue))
		return false;
	}
	return true;
    }
    
    // ------------------------------------------------------------
     /**
     * Sets a unique identifier (or change the existing one) in this
     * descriptor. The identifier is chosen arbitrarily.
     */
    public void setUniqueDocumentId() {
	StringGenerator strGen = StringGenerator.getGenerator();
	String docId = strGen.newString();
	setDocumentId(docId);
    }

    // ------------------------------------------------------------
    /**
     *
     * Produces a XML-like <code>String</code> representation of this
     * descriptor.
     * @return a <code>String</code> representation of this
     *         descriptor.
     */
    @Override
	public String toString() {

	String result = "<descriptor ";
	boolean first = true;
	for (String name: attributes_.keySet()) {
	    if (first) 
		first = false;
	    else
		result += ",";
	    String value = attributes_.get(name);
	    result += name + "=\"" + value + "\"";
	}
	result += ">";

	return result;
    }


    // ============================================================
    /**
     * @undocumented
     */
    public static void main(String[] args) throws Exception {

	Set<Descriptor> set = new HashSet<Descriptor>();
	
	Descriptor d1 = new Descriptor();
	d1.setAttribute("_docid", "d1");
	d1.setAttribute("att1", "fg");
	d1.setAttribute("att2", "pg mg");
	set.add(d1);
	System.out.println("d1=" + d1);

	Descriptor d2 = new Descriptor();
	d2.setAttribute("_docid", "d2");
	d2.setAttribute("att1", "ym");
	d2.setAttribute("att2", "pm mm");
	d2.setAttribute("att3", "mm sm");
	set.add(d2);
	System.out.println("d2=" + d2);

	Descriptor d3 = new Descriptor();
	d3.setAttribute("_docid", "d3");
	d3.setAttribute("att1", "nls");
	d3.setAttribute("att2", "pls mls");
	d3.setAttribute("att3", "als mls");
	set.add(d3);
	System.out.println("d3=" + d3);

	Descriptor p1 = new Descriptor();
	p1.setAttribute("att1", "(f|y).*");
	p1.setAttribute("att2", ".*(p|m).*");
	System.out.println("p1=" + p1);

	System.out.println("d1.matches(p1)=" + d1.matches(p1));
	System.out.println("d2.matches(p1)=" + d2.matches(p1));
	System.out.println("d3.matches(p1)=" + d3.matches(p1));

	System.out.println("Matching(d1, p1): " 
			   + d1.getMatchingAttributes(p1));
	System.out.println("Matching(d2, p1): " 
			   + d2.getMatchingAttributes(p1));
	System.out.println("Matching(d3, p1): " 
			   + d3.getMatchingAttributes(p1));

	System.out.println("matchingKeys: " + p1.getMatchingKeys(set));
        
        
        System.out.println("Attributes d1 = "+d1.asAttributes());
    }


}
