/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.queue;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.DescriptorZipBufferizer;
import casa.dodwan.message.MessageReceiver;
import casa.dodwan.pubsub.PubSubService;
import casa.dodwan.util.Bufferizer;
import casa.dodwan.util.Console;
import casa.dodwan.util.Service;

/**
 * DoDWAN queuing of messages. DoDWAN automatically creates a unique
 * instance of this class that should be accessed via a {@link
 * casa.dodwan.run.Dodwan} object. Message queues are registered with
 * DoDWAN specifying a pattern of subscription (descriptor that may
 * contain regular expressions) and a queue name. Messages that match
 * the pattern are stored in the queue, and they can be extracted from
 * the queue in a FIFO order. Message queues are persistent: if DoDWAN
 * is stopped and restarted, the queues are still available, together
 * with the messages they contain.
 */
public class QueueService 
    extends Service {
    
    private Console console_ = null;
    
    private PubSubService pubSubService_;

    Bufferizer<Descriptor> descriptorBufferizer_ = 
	new DescriptorZipBufferizer();

    File path_;

    private Map<String, MessageQueue> map_ = 
	new HashMap<String, MessageQueue>();


    // ------------------------------------------------------------
    public QueueService(PubSubService pubSubServ, 
			File path)
	throws Exception {
	
	super.suspend();
	
	pubSubService_ = pubSubServ;

	path_ = path;

	if (path_ == null)
	    return;

	if (!path_.exists())
	    path_.mkdirs();
	else
	    load(path_);
    }
    
    // ------------------------------------------------------------
    /**
     * Adds a message queue. 
     * @param key an identifier for the queue. 
     *        This identifier can be used for queue removing.
     * @param pattern a descriptor serving as a pattern for content-based reception.
     */
     public MessageQueue addQueue(String key, Descriptor pattern) {

	if (map_.containsKey(key)){
	    System.out.println("The key " + key + ", is already registered.");
	    return null;
	}

	if (verbosity.isEnabled())
	    System.out.println("QueueService.addQueue(" 
			       + key + ", "
			       + pattern + ")");
	
	casa.util.Logger.log("queue", "create " + key);
	MessageQueue queue;
	if (path_ == null) 
	    queue = new MessageQueue(null, pubSubService_);
	else {
	    File qPath = new File(path_, key);
	    queue = new MessageQueue(qPath, pubSubService_);
	    
	    try {
		descriptorBufferizer_.toFile(pattern, new File(qPath, "pattern"));
	    }
	    catch (Exception e) {
		System.out.println("QueueService.addQueue(): failed to record pattern in filesystem");
	    }
	}
	
	pubSubService_.addSubscription(key, pattern, queue);

	map_.put(key, queue);

	return queue;
     }

    // ------------------------------------------------------------
    private void rmdirs(File file) {

	if (! file.exists()) 
	    return;

	if (file.isDirectory()) {
	    File[] files = file.listFiles();
	    for (File f: files) 
		rmdirs(f);
	}
	file.delete();
    }

    // ------------------------------------------------------------
    /**
     * Removes a message queue.
     * Does nothing if the identifier <code>key</code> does not correspond to a 
     * registered queue.
     * @param key the identifier of the queue to be removed
     */
    public void removeQueue(String key) {
	
	if (verbosity.isEnabled())
	     System.out.println("QueueService.removeQueue("
				+ key + ")");

	if (! map_.containsKey(key)) {
	    System.out.println("QueueService.removeQueue("
			       + key + ") failed: this queue does not exist");
	    return;
	}

	pubSubService_.removeSubscription(key);
	
	if (path_ != null) {
	    File qPath = new File(path_, key);
	    rmdirs(qPath);
	}

	map_.remove(key);
	casa.util.Logger.log("queue", "remove " + key);
     }

      // ------------------------------------------------------------
      public Console console() {
 	  if (console_ == null)
 	      console_ = new QueueServiceConsole(this);
	return console_;
      }
    
    // ------------------------------------------------------------
    /**
     * Return the keys of the registered queues.
     * @return the set of the registered keys.
     */
    public Set<String> getKeys() {
	
	return map_.keySet();
    }
    
    // ------------------------------------------------------------
    public MessageQueue getQueue(String key) {
	
	return map_.get(key);
    }
    
    // ------------------------------------------------------------
    public Collection<MessageQueue> getQueues() {
	
	return map_.values();
    }
    
    // ------------------------------------------------------------
    public Descriptor getPattern(String key) {
	
	MessageReceiver rcv = pubSubService_.getReceiver(key);
	if (rcv != null)
	    return rcv.getPattern();
	else
	    return null;
    }
    
    // ------------------------------------------------------------
    public void load(File path) throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("QueueService.load(" + path.getAbsoluteFile() + ")");
	
	File[] files = path.listFiles();
	
	for (File file: files) {
	    if (! file.isDirectory()) 
		System.out.println("QueueService.load() error: "
				   + file.getAbsoluteFile()
				   + " is not a directory");
	    else {
		String key = file.getName();
		File pFile = new File(file, "pattern");
		if (! pFile.exists())
		    System.out.println("QueueService.load() error: "
				       + pFile.getAbsoluteFile()
				       + " is not available");
		else {
		    Descriptor pattern = descriptorBufferizer_.fromFile(pFile);
		    if (pattern != null)
			addQueue(key, pattern);
		}
	    }
	}
    }
    
    // ------------------------------------------------------------
    /**
      * Resumes this service. 
      * Indicates that this service is restarted after a suspend.
      * Presently nothing is done but printing out logging information 
      * according to the verbosity level.
      */
   @Override
   public void resume() {
       
	if (verbosity.isEnabled())
	    System.out.println("QueueService.resume()");
	
	if (isEnabled()) {
	    super.resume();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
     /**
      * Sets the verbosity level. 
      * @param level the level of verbosity. A value of 0 means that no information messages
      *        are printed out. A value different from 0 means that information messages are printed 
      *        out on standard output.
      */
    public void setVerbosity(int level) {
	
	switch (level) {
	case 0:
	    verbosity.disable();
	    break;
	default:
	    verbosity.enable();
	    break;
	}
	
	System.out.println("QueueService.setVerbosity(" + level + ")");
    }
    
    // ------------------------------------------------------------
     /**
      * Suspends this Queue service. 
      * Indicates that this service is temporarily suspended. It is typically  
      * called upon external events (host suspend to ram or suspend to disk).
      * Presently nothing is done but printing out logging information 
      * according to the verbosity level.
      */
    @Override
	public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("QueueService.suspend()");
	
	if (isEnabled()) {
	    super.suspend();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
     /**
     */
    public String toString() {

	return "QueueService"
	    + "\n nbQueues:" + String.valueOf(map_.size())
	    + "\n isEnabled:" + String.valueOf(isEnabled())
	    + "\n isSuspended:" + String.valueOf(isSuspended());
    }
    
    
}
