/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.queue;

import java.io.BufferedReader;
import java.io.PrintStream;
import java.util.Set;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.ArgumentParsingException;
import casa.dodwan.util.Command;
import casa.dodwan.util.Console;
import casa.dodwan.util.DataFormatting;


// import masc.metier.DocumentForwarder;

public class QueueServiceConsole extends Console {
    
    
    public QueueService queueService_;
    
    // ------------------------------------------------------------
    public QueueServiceConsole(QueueService queueService) {
	
	queueService_ = queueService;
	prompt_ = "queue% ";
    }
    
    // ============================================================
    // MAIN COMMANDS
    // ============================================================
    
    // ------------------------------------------------------------
    public QueueServiceConsole(String prompt, QueueService queueService) {
	
	queueService_ = queueService;
	prompt_ = prompt;
    }
    

    // ------------------------------------------------------------
    // add a message queue
    public void add(String[] arguments, PrintStream out) {
	
	String key = null;
	String descString = null;
	
	try {
	    key = ArgumentParsing.searchString("-k", arguments);
	    descString = ArgumentParsing.searchString("-desc", arguments);
	    
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}
	
	MessageQueue queue = queueService_.getQueue(key);
	if (queue != null) {
	    out.println("Sorry: a message queue with id '"
			+ key + "' already exists");
	    return;
	}

	Descriptor pattern = Descriptor.fromAttributes(descString);
	
	queueService_.addQueue(key, pattern);
    }
    
    // ------------------------------------------------------------
    public void disable(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	queueService_.disable();
    }
    
    // ------------------------------------------------------------
    public void enable(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	queueService_.enable();
    }
    
    // ------------------------------------------------------------
    @Override
	public void help(String offset, PrintStream out) {
	out
	    .println(offset + "[e]nable / [d]isable\n"
		     + offset + "[su]spend / [res]ume\n"
		     + offset + "[st]atus\n"
		     + offset + "[v]erbosity <level>\n"
		     + offset + "[a]dd [-k key] <-desc desc_string>\n"
		     + offset + "[rem]ove [keys...]>\n" 
		     + offset + "[g]et key\n"
		     + offset + "[m]onitor key timeout\n"
		     + offset + "[k]eys\n"
		     + offset + "[sh]ow <key>+");
    }
    
    // -------------------------------------------------------------
    public void processCommand(String[] arguments, BufferedReader reader,
			       PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("a")) {
	    add(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("rem")) {
	    remove(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("d")) {
	    disable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("e")) {
	    enable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("g")) {
	    get(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("h")) {
	    out.println("Pub/Sub Service: Command Set\n");
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("k")) {
	    keys(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("m")) {
	    monitor(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("res")) {
	    resume(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("sh")) {
	    show(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("st")) {
	    status(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("su")) {
	    suspend(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("v")) {
	    verbosity(arguments, out);
	    return;
	}
	
	out.println("Unknown command");
	help("  ", out);
    }
    
    // ------------------------------------------------------------
    public void remove(String[] arguments, PrintStream out) {
	
	if (arguments.length == 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("remove");
	
	for (int i = 1; i < arguments.length; i++) {
	    out.println("  " + arguments[i]);
	    queueService_.removeQueue(arguments[i]);
	}
    }
    
    // ------------------------------------------------------------
    public void resume(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	queueService_.resume();
    }
    
    // ------------------------------------------------------------
    public void keys(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("\nMessage queues:\n");
	Set keys = queueService_.getKeys();
	out.println(DataFormatting.toString(keys, "\n"));
    }

    // ------------------------------------------------------------
    public void get(String[] arguments, PrintStream out) {
	
	if (arguments.length != 2) {
	    help("  ", out);
	    return;
	}
	
	String key = arguments[1];
	
	MessageQueue queue = queueService_.getQueue(key);
	if (queue == null) {
	    out.println("Sorry: this queue does not exist");
	    return;
	}

	Message msg = queue.get();
	if (msg == null) {
	    out.println("Sorry: this queue is empty");
	    return;
	}

	out.println("Descriptor: " + msg.getDescriptor());
	out.println();
	out.println("Payload:");
	out.println(msg.getPayload());
    }

    // ------------------------------------------------------------
    public void monitor(String[] arguments, PrintStream out) {
	
	if (arguments.length != 3) {
	    help("  ", out);
	    return;
	}
	
	String key = arguments[1];
	long timeout = Long.parseLong(arguments[2]);
	
	MessageQueue queue = queueService_.getQueue(key);
	if (queue == null) {
	    out.println("Sorry: this queue does not exist");
	    return;
	}

	QueueConsumer qc = new QueueConsumer(queue, timeout);
	qc.start();
    }

    // ------------------------------------------------------------
    private void showQueue(String key, 
			   PrintStream out) {

	out.println("  key:     " + key);

	MessageQueue queue = queueService_.getQueue(key);
	if (queue == null)
	    out.println("  unknown...");
	else {			
	    out.println("  pattern: " + queueService_.getPattern(key));
	    out.println("  nbMsgs:  " + queue.nbMessages());
	}
    }

    // ------------------------------------------------------------
    public void show(String[] arguments, PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("show subscription(s)");
	
	if (arguments.length == 1) 
	    for (String key: queueService_.getKeys()) {
		showQueue(key, out);
	    }
	else
	    for (int i = 1; i < arguments.length; i++) {
		showQueue(arguments[i], out);
	    }
    }
    
    // ------------------------------------------------------------
    public void source(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	parseCommandFile(arguments[0], out);
    }
    
    // ============================================================
    // COMMAND PARSING
    // ============================================================
    
    // ------------------------------------------------------------
    public void status(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println(queueService_);
    }
    
    
    // ------------------------------------------------------------
    public void suspend(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	queueService_.suspend();
    }
    
    // ------------------------------------------------------------
    public void verbosity(String[] arguments, PrintStream out) {
	
	if (arguments.length != 2) {
	    help("  ", out);
	    return;
	}
	
	int level = -1;
	try {
	    level = Integer.parseInt(arguments[1]);
	} catch (Exception e) {
	    out.println("Could not parse verbosity level");
	}
	
	if (level >= 0)
	    queueService_.setVerbosity(level);
    }
    
}
