/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/




package casa.dodwan.run;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Locale;
import java.util.HashSet;
import java.util.Arrays;

import casa.dodwan.cache.CacheService;
import casa.dodwan.gossip.GossipingService;
import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.message.Message;
import casa.dodwan.pubsub.PubSubService;
import casa.dodwan.queue.QueueService;
import casa.dodwan.sendreceive.SendReceiveService;
import casa.dodwan.transmission.TransmissionService;
import casa.dodwan.util.Console;
import casa.util.Logger;
import casa.dodwan.util.Service;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Environment;
import casa.dodwan.util.LogManagement;

/**
 * DoDWAN main service. This class should be instantiated only once, unless it
 * is used for simulation purposes.
 */
public class Dodwan extends Service {

    /**
     * @undocumented
     */
    public Environment environment;

    /**
     * @undocumented
     */
    public CacheService cacheService;
    
    /**
     * @undocumented
     */
    public GossipingService gossipingService;
    
    /**
     * @undocumented
     */
    // public BeaconingService beaconingService;
    
    /**
     * @undocumented
     */
    private Console console_ = null;
    
    /**
     * Publication/subscription service. 
     */
    public PubSubService pubSubService;

    /**
     * @undocumented
     *
     * Send/Receive service. Provides access to direct content-based
     * communication, with no DTN.
     */
    // public SendReceiveService sendReceiveService;
    
    /**
     * @undocumented
     */
    public TransmissionService transmissionService;

    /**
     * @undocumented
     */
    public QueueService queueService;

    /**
     * @undocumented
     */
    public LogManagement logManagement;

    private Object lock_ = new Object();

    private boolean isOnline_ = false;

    // --------------------------------------------------------------------
    /**
     *  
     */
    public Dodwan()	throws Exception {

   	DodwanEnvironment env = DodwanEnvironment.getInstance();
	
	// Activity monitoring --------------------------
	long activityPeriod = env.activity_period;
	
	if (activityPeriod != 0) {
	    
	    File dir = new File(env.directory, "log");
	    String fname = SystemEnvironment.host + "_activity.log";
	    File logFile = new File(dir, fname);
	    
	    casa.dodwan.util.ActivityMonitor al = new casa.dodwan.util.ActivityMonitor(activityPeriod, logFile);
	    al.start();
	}
	
	// -----------------------------------------------
	
	String filter = System.getProperty("dodwan.log_filter");
	if (filter != null) {
	    String[] tags = filter.split(",");
	    Logger.filter(new HashSet<String>(Arrays.asList(tags)));
	}
	resetLogger();
	
	init(SystemEnvironment.host, env.directory);
    }

    // -------------------------------------------------------------------
    /**
     *  @undocumented
     *  Starts logging DoDWAN events in a new file. 
     */
    public static void resetLogger() {
	
	// System.out.println("Dodwan.resetLogger");
	resetLogger(null);
    }

    // -------------------------------------------------------------------
    /**
     *  @undocumented
     *  Starts logging DoDWAN events in the specified file. 
     */
    public static void resetLogger(String prefix) {
	
	// System.out.println("Dodwan.resetLogger");
	
	SimpleDateFormat dateFormat = 
	    new SimpleDateFormat("yyyy-MM-dd-HH-mm", Locale.ENGLISH);
	long now = System.currentTimeMillis();
	String fname = (prefix==null?"":prefix + "_")
	    + SystemEnvironment.host 
	    + "_" + dateFormat.format(now) + ".log";

	File dir = new File(DodwanEnvironment.getInstance().directory, "log");
	File logFile = new File(dir, fname);

	Logger.enable(logFile, true, 
		      SystemEnvironment.host);

	System.out.println("Now logging in file "
			   + logFile);
    }

    // ---------------------------------------------------------------------
    /**
     * @undocumented
     */
    public Console console() {
	if (console_ == null)
	    console_ = new DodwanConsole(this);
	return console_;
    }
 
   
    // ----------------------------------------------------------------------
    private void init(String host, 
		      File dir)
	throws Exception {

	Logger.log("dodwan", "seed=" + casa.dodwan.util.RandomSeed.getSeed());
	
	// Set default StringGenerator so that all strings generated
	// will use the local host name as a prefix
	casa.dodwan.util.StringGenerator.setSeed(host);

	// -------------------------------------------------
	// CACHE SERVICE
	cacheService = 
	    new CacheService((dir == null ? null : new File(dir, "cache")));

	// ------------------------------------------------------------
	// ENVIRONMENT MANAGEMENT
	environment = new Environment();

	// -------------------------------------------------
	// TRANSMISSION SERVICE
	transmissionService = new TransmissionService(environment);
	
	// -------------------------------------------------
	// GOSSIPING SERVICE
	gossipingService = 
	   new GossipingService(transmissionService,
				 cacheService.cache, environment);
	transmissionService.dispatcher.addSink(gossipingService);
	// environment.addListener(transmissionService.channelManager);

	// --------------------------------------------------
	// PUB/SUB SERVICE
	pubSubService = 
	   new PubSubService(cacheService,
			     gossipingService,
			     (dir == null ? null : new File(dir, "pubsub")));
	// transmissionService.dispatcher.addSink(pubSubService);

	// --------------------------------------------------
	// QUEUE SERVICE
	queueService = 
	    new QueueService(pubSubService,
			     (dir == null ? null : new File(dir, "queue")));

	// --------------------------------------------------
	// SEND/RECEIVE SERVICE
	// sendReceiveService =
	//     new SendReceiveService(transmissionService.loop);
	// transmissionService.dispatcher.addSink(sendReceiveService);
	
//  	casa.dodwan.docware.Descriptor pattern = new casa.dodwan.docware.Descriptor();
//  	pattern.setAttribute("_id", ".*");
//  	sendReceiveService.addSubscription("verbose",p attern,
//  					   new casa.dodwan.util.VerboseProcessor<Message>());

	// --------------------------------------------------
	// BEACONING SERVICE
	// FG: suppression service de beaconing
	// beaconingService = new BeaconingService(transmissionService.channelManager,
	// 					environment);
	// transmissionService.dispatcher.addSink(beaconingService);

	
	// --------------------------------------------------
	// ENABLING SERVICES
	transmissionService.enable();
	boolean gossiping_enabled =
	    DodwanEnvironment.getInstance().gossiping_enabled;
	Logger.log("dodwan", "gossiping_enabled=" + gossiping_enabled);
		   
	if (gossiping_enabled) {
	    
	    cacheService.enable();
	    gossipingService.enable();
	    pubSubService.enable();
	}
	// sendReceiveService.enable();
	// beaconingService.enable();
	enable();

	
	console_ = new DodwanConsole(this);

	resume();
	
	logManagement = 
	    new LogManagement(this, new File(dir, "log"));
    }


    // -----------------------------------------------------------------------
    /**
     * Tells DoDWAN to perform the necessary tasks after the system has resumed
     * from a suspend state (typically after a suspend to RAM). This method
     * should be called upon an external event indicating that the system
     * returns from a suspend state.
     */
    @Override
	public void resume() {
	
	if (verbosity.isEnabled())
	    System.out.println("Dodwan.resume()");
	
	super.resume();

	if (! Logger.isEnabled())
	    resetLogger();
	
	Logger.log("dodwan", "resume");

	if (DodwanEnvironment.getInstance().gossiping_enabled)
	    pubSubService.resume();
    }
    
    // ----------------------------------------------------------------------
    /**
     * Tells DoDWAN to perform the necessary tasks before the system enters a
     * suspen mode (typically when the system is suspended to RAM). This method
     * should be called upon an external event indicating that the system
     * enters a suspend state.
     */
    @Override
	public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("Dodwan.suspend()");
	
	super.suspend();

	pubSubService.suspend();
	// sendReceiveService.suspend();
	// beaconingService.suspend();
	gossipingService.suspend();
	cacheService.suspend();
	transmissionService.suspend();
	environment.clearPeers();

	if (Logger.isEnabled()) {
	    Logger.log("dodwan", "suspend");
	    Logger.disable();
	}
    }


    
    // -----------------------------------------------------------------------
    /**
     * Returns the current status (online/offline) of DoDWAN.
     */
	public boolean isOnline() {
	    return isOnline_;
	}
    
    // -----------------------------------------------------------------------
    /**
     * Tells DoDWAN to perform the necessary tasks after the system
     * has gone online.
     */
	public void online() {
	
	    if (isOnline_)
		return;

	if (verbosity.isEnabled())
	    System.out.println("Dodwan.online()");
	
	Logger.log("dodwan", "online");

	transmissionService.resume();
	try {
	    Thread.sleep(200);
	}
	catch (Exception e) {}
	if (DodwanEnvironment.getInstance().gossiping_enabled) {
	    cacheService.resume();
	    gossipingService.resume();
	}
	// FG beaconingService.resume();
	// sendReceiveService.resume();

	isOnline_ = true;
    }
    
    // ----------------------------------------------------------------------
    /**
     * Tells DoDWAN to perform the necessary tasks before the system
     * goes offline. Once offline, all transmissions are disabled,
     * although functions that do not require any network access are
     * still available (for example, publishing in the local cache).
     */
    public void offline() {
	
	if (! isOnline_)
	    return;

	if (verbosity.isEnabled())
	    System.out.println("Dodwan.offline()");
	
	// sendReceiveService.suspend();
	Logger.log("dodwan", "OFF1");
	gossipingService.suspend();
	// beaconingService.suspend();
	// cacheService.suspend();
	Logger.log("dodwan", "OFF2");
	transmissionService.suspend();
	Logger.log("dodwan", "OFF3");
	environment.clearPeers();

	Logger.log("dodwan", "offline");
	isOnline_ = false;
    }



    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void doWait(){
	synchronized(lock_){
	    try {
		lock_.wait();
	    } 
	    catch(InterruptedException e) {
		System.err.println("Dodwan.doWait() failed");
	    }
	}
    }



    // ------------------------------------------------------------
    /**
     * @undocumented
     */
    public void doNotify(){
	synchronized(lock_){
	    lock_.notify();
	}
    }

 
}
