/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.transmission;

import java.util.Map;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.NetworkInterface;

import casa.dodwan.net.UDP_Channel;

import casa.dodwan.config.DodwanEnvironment;
// import casa.dodwan.forwarding.ForwardingProtocol;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
// import casa.dodwan.net.InetSocket;
// import casa.dodwan.net.MulticastInetSocket;
import casa.dodwan.util.Console;
import casa.dodwan.util.Dispatcher;
import casa.dodwan.util.ListenableSink;
import casa.dodwan.util.Service;
import casa.dodwan.util.Sink;
import casa.util.Processor;
import casa.dodwan.util.Socket;
import casa.dodwan.util.SuspendableSink;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.Environment;
import casa.util.Logger;
import casa.dodwan.net.Channel;
import casa.dodwan.net.ChannelManager;


public class TransmissionService extends Service
    implements Sink<Message>, Processor<Message> {


    private Console console_ = null;
    
    public Dispatcher<Message> dispatcher;
    
    public Dispatcher<Message> loop;
    
    public SuspendableSink<Message> outputSink;
    public SuspendableSink<Message> inputSink;
    
    // public ForwardingProtocol protocol = null;
    
    public ChannelManager channelManager;
    
    public Environment environment;

    private String hostId_;
    
    // ------------------------------------------------------------
    public TransmissionService(Environment envt) 
	throws Exception {

	environment = envt;
	
	dispatcher = new Dispatcher<Message>();

	inputSink = new SuspendableSink<Message>(dispatcher, true);

	hostId_ = SystemEnvironment.host;
	
	channelManager =
	    new ChannelManager(environment, this);
	
	outputSink = new SuspendableSink<Message>(this, true);
	
	// ------------------------------------------------------------
	// TRANSMISSION LOOP
	loop = new Dispatcher<Message>();
	loop.addSink(outputSink);
	loop.addSink(dispatcher);
	
    }

    // ------------------------------------------------------------
    public Console console() {
	if (console_ == null)
	    console_ = new TransmissionServiceConsole(this);
	return console_;
    }
    
    // ------------------------------------------------------------
    public void disable() {
	
	if (verbosity.isEnabled())
	    System.out.println("TransmissionService.disable()");
	
	if (isEnabled()) {
	    outputSink.disable();
	    inputSink.disable();
	    super.disable();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is not enabled");
    }
    
    // ------------------------------------------------------------
    public void enable() {
	
	if (verbosity.isEnabled())
	    System.out.println("TransmissionService.enable()");
	
	if (!isEnabled()) {
	    outputSink.enable();
	    inputSink.enable();
	    super.enable();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is aleady enabled");
    }
    
    // ------------------------------------------------------------
    public void resume() {
	
	if (verbosity.isEnabled())
	    System.out.println("TransmissionService.resume()");
	
	if (isEnabled()) {
	    for (Channel channel: channelManager.getChannels().values())
		if (! channel.isOpen())
		    channel.open();
		    
	    outputSink.resume();
	    inputSink.resume();
	    super.resume();
	}
    else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
    public void setVerbosity(int level) {
	
	switch (level) {
	case 0:
	    verbosity.disable();
	    dispatcher.verbosity.disable();
	    loop.verbosity.disable();
	    break;
	default:
	    verbosity.enable();
	    dispatcher.verbosity.enable();
	    loop.verbosity.enable();
	    break;
	}
    }
    
    // ------------------------------------------------------------
    public void suspend() {
	
	if (verbosity.isEnabled())
	    System.out.println("TransmissionService.suspend()");
	
	if (isEnabled()) {
	    channelManager.clear();
	    outputSink.suspend();
	    inputSink.suspend();
	    super.suspend();
	}
	else if (verbosity.isEnabled())
	    System.out.println("  Failure: this service is disabled");
    }
    
    // ------------------------------------------------------------
    public String toString() {

	return "TransmissionService"
	    + "\n isEnabled: " + String.valueOf(isEnabled())
	    + "\n isSuspended:" + String.valueOf(isSuspended())
	    + "\n threshold:" + String.valueOf(DodwanEnvironment.getInstance().fragmentation_threshold);
    }

    // ============================================================
    // Implementation of the Sink<Message> interface
    // ============================================================
    
    private void logTransmission(Message msg) {
	
	// FG: debug
	Descriptor desc = msg.getDescriptor();
	String sdr = desc.getAttribute("_sdr");
	String rcv = desc.getAttribute("_rcv");
	if (rcv != null)
	    Logger.log("debug_trans", sdr + " > " + rcv + " [" + desc.getDocumentId()
		       + "]");
    }
    
    // ------------------------------------------------------------
    public void process(Message msg) {

	// logTransmission(msg);
	
	try {
	    inputSink.write(msg);
	}
	catch (Exception e) {
	    Logger.log("trans", "warning: could not forward message to input sink");
	}
    }

    // ------------------------------------------------------------
    public void write(Message msg) throws Exception {
	
	long now = System.currentTimeMillis();

	Descriptor desc = msg.getDescriptor();
	desc.setAttribute("_sdr", hostId_);
	desc.setAttribute("_st", Long.toString(now));
	String rcv = desc.getAttribute("_rcv");

	// String key = desc.getDocumentId();
	// String type = desc.getServiceType();
	// Logger.log("chan_mgr", 
	// 	   ">[" + msg.getPayload().length
	// 	   + "] id=" + key
	// 	   + ",rcv=" + rcv
	// 	   + ",type=" + type
	// 	   + ",st=" + now);
	
	send(msg, rcv);
    }

    // ------------------------------------------------------------
    private void send(Message msg, String rcv) 
	throws Exception {

	// logTransmission(msg);
	
	// If the receiver is not specified, there is nothing we
	// can do here...
	if (rcv == null) {
	    Logger.log("trans", "> sending to all UDP channels (rcv==null)");
	    Map<String,Channel> channels = channelManager.getChannels();
	    for (Channel chan: channels.values()) {
		if (chan instanceof UDP_Channel) 
		    chan.write(msg);
	    }
	    return;
	}
	
	try {
	    // Otherwise let us check that a Channel is available
	    // to reach the set receiver
	    Channel channel = channelManager.getChannelForPeer(rcv);
	    
	    if ((channel != null) && channel.isOpen()) {
		channel.write(msg);
	    }
	    else
		Logger.log("trans", "> failed (no channel handler for rcv="
			   + rcv + ")");
	
	} catch(Exception e) {
	    Logger.log("trans", "> failed");
	    throw e;
	}
    }


}
