/* *****************************************************************************
 * Copyright (c) 2005-2010 VALORIA Laboratory, 
 * Universite Europeenne de Bretagne, Universite de Bretagne-Sud, France
 * <http://www-valoria.univ-ubs/CASA/DoDWAN>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************
 * $Id: Log2DGS.java $
 * ****************************************************************************/

package casa.dodwan.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
    
/**
 * Parses a log file and detects all events pertaining to node
 * activity (i.e. online/offline events) as well as neighbor discovery
 * (i.e. add/remove neighbor). Produces a DGS file based on these
 * events.
 */
public class Log2DGS extends LogParser {

    private long lastTimeDisplayed_ = -1;
    private long period = 1;
    
    private Map<String,Set<String>> nodes_ =
	new HashMap<String,Set<String>>();

    // ------------------------------------------------------------
    public Log2DGS(InputStream in, OutputStream out) {
	super(in, out);
    }

    // ------------------------------------------------------------
    public void setPeriod(long period) {
	
	this.period = period;
    }

    // ------------------------------------------------------------
    public void processEvent(Event event) {

	if (event.tag.equals("dodwan")
	    && event.message.startsWith("online")) {
	    processOnline(event);
	    return;
	}

	if (event.tag.equals("dodwan")
	    && event.message.startsWith("offline")) {
	    processOffline(event);
	    return;
	}

	if (event.tag.equals("gossip")
	    && event.message.startsWith("add")) {
	    processAddNeighbor(event);
	    return;
	}

    	if (event.tag.equals("gossip")
	    && event.message.startsWith("remove")) {
	    processRemoveNeighbor(event);
	    return;
	}
    }

    // ------------------------------------------------------------
    private void displayTime(long now) {

	now = period * (now / period);
	
	if (now != lastTimeDisplayed_) {
	    lastTimeDisplayed_ = now;
	    System.out.println("st " + now);
	}
    }
    
    // ------------------------------------------------------------
    private String getEdgeId(String id1, String id2) {

	if (id1.compareTo(id2) <= 0)
	    return id1 + "-" + id2;
	else
	    return id2 + "-" + id1;
    }

    // ------------------------------------------------------------
    private boolean isOnline(String id) {

	Set<String> neigh = nodes_.get(id);
	return (neigh != null);
    }

    // ------------------------------------------------------------
    private void processOnline(Event event) {

	if (isOnline(event.name)) {
	    System.err.println("Error: node " + event.name + " is already online");
	    System.exit(1);
	}
    
	addNode(event.name);
	displayTime(event.time);
	System.out.println("an " + event.name);
    }
    

    // ------------------------------------------------------------
    private void addNode(String id) {

	nodes_.put(id, new HashSet<String>());
    }

    // ------------------------------------------------------------
    private void processOffline(Event event) {

	if (! isOnline(event.name)) {
	    System.err.println("Error: node " + event.name + " is not online");
	    System.exit(1);
	}
	       
	displayTime(event.time);
	deleteNode(event.name);

	System.out.println("dn " + event.name);
    }
    

    // ------------------------------------------------------------
    private void deleteNode(String id) {

	Set<String> neigh = nodes_.get(id);
	if (! neigh.isEmpty()) {
	    System.out.println("# Removing edges before node " + id + " is removed");
	    Set<String> tmp = new HashSet<String>();
	    tmp.addAll(neigh);
	    for (String id2: tmp) {
		deleteEdge(id, id2);
	    }
	}

	nodes_.remove(id);
    }

    // ------------------------------------------------------------
    private boolean areNeighbors(String node1, String node2) {

	Set<String> neigh1 = nodes_.get(node1);
	Set<String> neigh2 = nodes_.get(node2);

	if ((neigh1 == null) || (neigh2 == null))
	    return false;
	
	return (neigh1.contains(node2));
    }

    // ------------------------------------------------------------
    private void processAddNeighbor(Event event) {

	String[] fields = event.message.split(" ");
	String node1 = event.name;
	String node2 = fields[1];

	if (! (isOnline(node1) && isOnline(node2)))
	    return;
	   
	if (areNeighbors(node1, node2))
	    return;

	displayTime(event.time);
	addEdge(node1, node2);
    }

    // ------------------------------------------------------------
    private void addEdge(String node1, String node2) {

	if (areNeighbors(node1, node2))
	    return;
	
	if (node1.compareTo(node2) < 0)
	    System.out.println("ae " + node1 + "-" + node2
			       + " " + node1 + " " + node2);
	else
	    System.out.println("ae " + node2 + "-" + node1
			       + " " + node2 + " " + node1);

	Set<String> neigh = nodes_.get(node1);
	neigh.add(node2);
	neigh = nodes_.get(node2);
	neigh.add(node1);
    }
    
    // ------------------------------------------------------------
    private void processRemoveNeighbor(Event event) {

	String[] fields = event.message.split(" ");
	String node1 = event.name;
	String node2 = fields[1];

	if (! areNeighbors(node1, node2))
	    return;
	
	displayTime(event.time);
	deleteEdge(node1, node2);
    }

    // ------------------------------------------------------------
    private void deleteEdge(String node1, String node2) {

	if (node1.compareTo(node2) < 0)
	    System.out.println("de " + node1 + "-" + node2);
	else
	    System.out.println("de " + node2 + "-" + node1);
	
	Set<String> neigh = nodes_.get(node1);
	neigh.remove(node2);
	neigh = nodes_.get(node2);
	neigh.remove(node1);
    }
    
    // ------------------------------------------------------------
    private void close() {

	Set<String> tmp = new HashSet<String>();

	if (tmp.isEmpty())
	    return;
	
	System.out.println("# Removing all remaining nodes");
	tmp.addAll(nodes_.keySet());
	for (String id: tmp) {
	    deleteNode(id);
	}
    }

    // ------------------------------------------------------------
    public static void usage() {
        System.out.println("java Log2DGS [ref_time] [period] [-|logfile]");
        System.out.println("Reads a log file on stdin and produces a DGS file based on log events");
	System.out.println("pertaining to node activity (i.e. online|offline) and neighbor");
	System.out.println("discovery (i.e. add/remove).");
        System.exit(0);
    }

   // ------------------------------------------------------------
    public static void main(String[] args) throws IOException {
        if ((args.length > 0 && args[0].equals("-h"))
	    || (args.length != 3)) {
            usage();
        }

	long ref_time = Long.parseLong(args[0]);
	long period   = Long.parseLong(args[1]);
	
	InputStream is;
	if (args[1].equals("-"))
	    is = System.in;
	else
	    is = new FileInputStream(args[2]);
	
        Log2DGS parser =
	    new Log2DGS(is, System.out);
	parser.setRefTime(ref_time);
	parser.setPeriod(period);
	
	System.out.println("DGS004");
	System.out.println("null 0 0");
	System.out.println("cg ref_time=\"" + ref_time + "\"");
	
	parser.parse();
	parser.close();
	
    }
}
