/* *****************************************************************************
 * Copyright (c) 2005-2010 VALORIA Laboratory, 
 * Universite Europeenne de Bretagne, Universite de Bretagne-Sud, France
 * <http://www-valoria.univ-ubs/CASA/DoDWAN>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************
 * $Id: Log2DGS.java $
 * ****************************************************************************/

package casa.dodwan.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.FileInputStream;
import java.io.OutputStream;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
    
/**
 * Parses a log file and detects all events pertaining to node
 * activity (i.e. online/offline events) as well as neighbor discovery
 * (i.e. add/remove neighbor) and reeive events. Produces a file
 * describing these events.
 */
public class Log2Events extends LogParser {

    // ------------------------------------------------------------
    public Log2Events(InputStream in, OutputStream out) {
	super(in, out);
    }

    // ------------------------------------------------------------
    private String getFields(String[] fields,
			     int idx1, int idx2) {

	String result = null;
	if (idx2 >= fields.length)
	    idx2 = fields.length - 1;
	for (int i=idx1; i <= idx2; i++) {
	    if (i == idx1)
		result = fields[i];
	    else
		result += " " + fields[i];
	}

	return result;
    }

    // ------------------------------------------------------------
    public void processEvent(Event event) {

	String[] fields = event.message.split(" +");
	
	if (event.tag.equals("dodwan")) {
	    if (fields[0].equals("online")) {
		display(event.time, "online", event.name, null);
		return;
	    }

	    if (fields[0].equals("offline")) {
		display(event.time, "offline", event.name, null);
		return;
	    }
	}

	if (event.tag.equals("gossip")) {
	    if (fields[0].equals("add")) {
		display(event.time, "cx-on", event.name, "peer=" + fields[1]);
		return;
	    }

	    if (fields[0].equals("remove")) {
		display(event.time, "cx-off", event.name, "peer=" + fields[1]);
		return;
	    }
	}

	if ((event.tag.equals("offer")
	     || event.tag.equals("request")
	     || event.tag.equals("profile"))   
	    && (fields[0].equals("<"))) {
	    int idx1 = fields[1].indexOf("=");
	    int idx2 = fields[1].indexOf(",", idx1);
	    String sdr = fields[1].substring(idx1+1, idx2);
	    display(event.time, event.tag, event.name, "sdr=" + sdr);
	    return;
	}

	if (event.tag.equals("provide")
	    && fields[0].equals("<")) {

	    String last_field = fields[fields.length-1];
	    if ((! last_field.equals("(dup)"))
		&& (! last_field.equals("(alien)"))) {
		int idx1 = fields[1].indexOf("=");
		int idx2 = fields[1].indexOf(",", idx1);
		String sdr = fields[1].substring(idx1+1, idx2);
		idx1 = fields[1].indexOf("=", idx2);
		idx2 = fields[1].indexOf("/");
		String mid = fields[1].substring(idx1 + 1, idx2);
		// Let us ignore message fragments other than fragment 0
		if (mid.endsWith("/0"))
		    return;
		// display(event.time, "provide", event.name,
		// 	"sdr=" + sdr + " mid=" + mid);
		display(event.time, "rcv", event.name,
			"sdr=" + sdr + " mid=" + mid);
		return;
	    }
	}
	
	if (event.tag.equals("udp_chan")
	    && fields[0].equals("<")) {

	    if (fields[1].equals("(hello)")) {
		int idx1 = fields[2].indexOf("=");
		int idx2 = fields[2].indexOf(",", idx1);
		String sdr = fields[2].substring(idx1+1, idx2);
		display(event.time, "hello", event.name, "sdr=" + sdr);
		return;
	    }

	    if (fields[1].equals("(bye)")) {
		int idx1 = fields[2].indexOf("=");
		String sdr = fields[2].substring(idx1+1);
		display(event.time, "bye", event.name, "sdr=" + sdr);
		return;
	    }
	    return;
	}

	if (event.tag.equals("pubsub")
	    && fields[0].equals(">")
	    && fields[1].startsWith("publish")) {

	    int idx1 = fields[1].indexOf("=");
	    int idx2 = fields[1].indexOf(")");
	    String mid = fields[1].substring(idx1 + 1, idx2);
	    display(event.time, "snd", event.name, "mid=" + mid);

	    return;
	}

	if (event.tag.equals("bcast")) {

	    if (fields[0].equals("deliver")) {
		display(event.time, "dlv", event.name, 
			getFields(fields, 1, 5));
		return;
	    }

	    if (fields[0].equals("expired")) {
		display(event.time, "exp", event.name, 
			getFields(fields, 1, 3));
		return;
	    }

	    if (fields[0].equals("publish")) {
		display(event.time, "pub", event.name,
			getFields(fields, 1, 5));
		return;
	    }
	    
	    if (fields[0].equals("rcvd")) {
		display(event.time, "rcvd", event.name, 
			getFields(fields, 1, 3));
		return;
	    }

	    if (fields[0].equals("status")) {
		display(event.time, "stat", event.name, 
			getFields(fields, 1, 7));
		return;
	    }
	}

	if (event.tag.equals("crdt")
	    && fields[0].equals("<")) {
		
	    if (fields[1].indexOf("digest") > -1) {
		int idx1 = fields[1].indexOf("=");
		int idx2 = fields[1].indexOf(",", idx1);
		String sdr = fields[1].substring(idx1+1, idx2);
		display(event.time, "digest", event.name, "sdr=" + sdr);
		return;
	    }

	    if (fields[1].indexOf("delta") > -1) {
		int idx1 = fields[1].indexOf("src=");
		int idx2 = fields[1].indexOf(",", idx1);
		String sdr = fields[1].substring(idx1+4, idx2);
		idx1 = fields[1].indexOf("size=");
		idx2 = fields[1].indexOf(",", idx1);
		String size = fields[1].substring(idx1+5);
		display(event.time, "delta", event.name,
			"sdr=" + sdr + " size=" + size);
		return;
	    }

	    if (fields[1].indexOf("full") > -1) {
		int idx1 = fields[1].indexOf("=");
		int idx2 = fields[1].indexOf(",", idx1);
		String sdr = fields[1].substring(idx1+1, idx2);
		idx1 = fields[1].indexOf("size=");
		idx2 = fields[1].indexOf(",", idx1);
		String size = fields[1].substring(idx1+5);
		display(event.time, "full", event.name,
			"sdr=" + sdr
			+ " size=" + size);
		return;
	    }
	}

	if (event.tag.equals("crdt_tester")
	    && fields[0].startsWith("val")) {

	    int idx1 = fields[0].indexOf("(");
	    int idx2 = fields[0].indexOf(")");
	    int idx3 = fields[0].indexOf("=");

	    String id = fields[0].substring(idx1 + 1, idx2);
	    String val = fields[0].substring(idx3 + 1);
	    
	    display(event.time, "status", event.name,
		    "id=" + id + " val=" + val);
	    return;
	}
    }

    // ------------------------------------------------------------
    private void display(long time, String type, String id, String args) {

	System.out.println(time + " " + type + " " + id
			   + (args==null?"":" " + args));
    }

    // ------------------------------------------------------------
    private String getEdgeId(String id1, String id2) {

	if (id1.compareTo(id2) <= 0)
	    return id1 + "-" + id2 + " dir=<";
	else
	    return id2 + "-" + id1 + " dir=>";
    }

   // ------------------------------------------------------------
    public static void usage() {
        System.out.println("java Log2Events [ref_time] [-|logfile]");
        System.out.println("Reads a log file on stdin and produces on stdout a file that lists all events");
	System.out.println("detected in the log file");
        System.exit(0);
    }

   // ------------------------------------------------------------
    public static void main(String[] args) throws IOException {
        if ((args.length > 0 && args[0].equals("-h"))
	    || (args.length != 2)) {
            usage();
        }

	long ref_time = Long.parseLong(args[0]);
	
	InputStream is;
	if (args[1].equals("-"))
	    is = System.in;
	else
	    is = new FileInputStream(args[1]);
	
        Log2Events parser =
	    new Log2Events(is, System.out);
	parser.setRefTime(ref_time);
	
	parser.parse();
	
    }
}
