/* *****************************************************************************
 * Copyright (c) 2005-2010 VALORIA Laboratory, 
 * Universite Europeenne de Bretagne, Universite de Bretagne-Sud, France
 * <http://www-valoria.univ-ubs/CASA/DoDWAN>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************
 * $Id: LogParser.java 3919 2012-05-20 15:08:16Z guidec $
 * ****************************************************************************/

package casa.dodwan.util;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.BufferedReader;
import java.io.PrintStream;
import java.text.SimpleDateFormat;
import java.util.Locale;

/**
 * Provides facilities for parsing a log file, as produced by
 * casa.dodwan.util.Logger.  
 *
 * Event messages must be of the form <em>date</em> <em>name</em>
 * <em>tag</em> <em>message</em>. Dates must be written with the
 * format <em>yyyy-MM-dd HH:mm:ss.SSS</em>.
 */
public class LogParser {

    private BufferedReader in;
    protected PrintStream out;
    protected long ref_time = 0;
    
    private static final SimpleDateFormat dateFormat = 
	new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS", Locale.ENGLISH);

    // ==================================================
    protected class Event {

	long time;
	String name;
	String tag;
	String message;
	
	public Event(long time, String name,
		     String tag, String message) {

	    this.time = time; this.name = name;
	    this.tag = tag; this.message = message;
	}

	public String toString() {

	    return dateFormat.format(time)
		+ " " + name
		+ " " + tag + ": " + message;
	}
    }
    // ==================================================
    

    public LogParser(InputStream in, OutputStream out) {
	
        this.in = new BufferedReader(new InputStreamReader(in));
        this.out = new PrintStream(out);
    }

    // ------------------------------------------------------------
    public void setRefTime(long ref_time) {
	
	this.ref_time = ref_time;
    }

    // ------------------------------------------------------------
    public long getRefTime() {
	
	return this.ref_time;
    }

    // ------------------------------------------------------------
    public void processEvent(Event event) {
	
	out.println(event);
    }

    // ------------------------------------------------------------
    private long parseTime(String str) throws Exception {

	int idx = str.indexOf(":");
	if (idx != -1)
	    return dateFormat.parse(str).getTime();
	else
	    return Long.parseLong(str);
    }

    // ------------------------------------------------------------
    private Event parseEvent(String str)
	throws Exception {

	int idx1 = str.indexOf(" ");
	idx1 = str.indexOf(" ", idx1+1);
	String str_time = str.substring(0, idx1).trim();
	long time = parseTime(str_time) - ref_time;
	int idx2 = str.indexOf(" ", idx1+1);
	String name = str.substring(idx1+1, idx2).trim();
	int idx3 = str.indexOf(":", idx2+1);
	String tag = str.substring(idx2+1, idx3).trim();
	String message = str.substring(idx3+1).trim();

	return new Event(time, name, tag, message);
    }

    // ------------------------------------------------------------
    public void parse()
	throws IOException {

	String line = null;
	int cnt = 0;
	try {
	    line = in.readLine(); cnt++;
	    while (line != null) {
		Event event = parseEvent(line);
		processEvent(event);
		line = in.readLine(); cnt++;
	    }
	}
	catch (Exception e) {
	    System.err.println("Failed while parsing line " + cnt + ": " + line);
	    e.printStackTrace();
	}
	
        out.flush();
    }

    // ------------------------------------------------------------
    public static void usage() {
        System.out.println("java LogParser");
        System.out.println("Reads a log file on stdin and produces something on stdout.");
        System.exit(0);
    }

    // ------------------------------------------------------------
    public static void main(String[] args) throws IOException {
        if (args.length > 0 && args[0].equals("-h")) {
            usage();
        }

        LogParser parser = new LogParser(System.in, System.out);	
	parser.parse();
    }
}
