/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/


package casa.dodwan.util;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.nio.charset.Charset;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;

public class MimeUtility {
	
	private static final char CR = (char)13, LF = (char)10, SPACE = (char)32, HTAB = (char)9;
	private static final int splitLinesAt = 60;
	private static final char[] base64code = {
		'A','B','C','D','E','F','G','H','I','J','K','L','M','N','O','P','Q','R','S','T','U','V','W','X','Y','Z',
		'a','b','c','d','e','f','g','h','i','j','k','l','m','n','o','p','q','r','s','t','u','v','w','x','y','z',
		'0','1','2','3','4','5','6','7','8','9','+','/'};
	private static final char[] hexdigits = {'0','1','2','3','4','5','6','7','8','9','A','B','C','D','E','F'};	
	
	/**
	 * Decode "unstructured" headers, that is, headers that are defined as '*text' as per RFC 822.
	 * 
	 * The string is decoded using the algorithm specified in RFC 2047, Section 6.1. If the 
	 * charset-conversion fails for any sequence, an UnsupportedEncodingException is thrown. 
	 * If the String is not an RFC 2047 style encoded header, it is returned as-is 
	 * 
	 * @param etext the possibly encoded value 
	 * @return
	 * @throws ParseException if the string is not an encoded-word as per RFC 2047 and RFC 2231.
	 * @throws UnsupportedEncodingException if the charset conversion failed.
	 */
	public static String decodeText(String etext) throws ParseException, UnsupportedEncodingException {
		StringBuffer buffer = new StringBuffer();
		boolean prevucd = false; // previous word uncoded
		for (String word : split(etext)) {
			if (prevucd)
				buffer.append(" ");
			// decode word if is encoded - else do nothing (ascii)
			if (isEncodedWord(word)) {
				word = decodeWord(word);
				buffer.append(word);
				prevucd = false;
			} else {
				buffer.append(word);
				prevucd = true;
			}
		}
		return buffer.toString();
	}

	/**
	 * Encode a RFC 822 "text" token into mail-safe form as per RFC 2047. 
	 * 
	 * The given Unicode string is examined for non US-ASCII characters. If the 
	 * string contains only US-ASCII characters, it is returned as-is. If the string 
	 * contains non US-ASCII characters, it is first character-encoded using the specified 
	 * charset, then transfer-encoded using either the B or Q encoding. The resulting bytes 
	 * are then returned as a Unicode string containing only ASCII characters. 
	 * 
	 * @param text Unicode string 
	 * @param charset the charset. If this parameter is null, the platform's default chatset is used.
	 * @param encoding the encoding to be used. Currently supported values are "B" and "Q". 
	 * If this parameter is null, then the "Q" encoding is used if most of characters to be 
	 * encoded are in the ASCII charset, otherwise "B" encoding is used. 
	 * @return Unicode string containing only US-ASCII characters 
	 */
	public static String encodeText(String text, String charset, String encoding) {
		if (text.length() == 0)
			return text;
		
		if (charset == null)
			charset = Charset.defaultCharset().toString();
		
		// counts the ascii characters in the given text
		int nbascii = 0;
		for (int i=0; i<text.length(); i++)
			if (text.charAt(i) < 128)
				nbascii++;
		if (nbascii == text.length()) encoding = "NONE";
		else if (encoding == null) {
			if (nbascii <= text.length() / 2) encoding = "B";
			else encoding = "Q";
		}

		// decide which encoding will be applied
		if (encoding.equals("NONE")) return text; // NONE
		else if (encoding.equalsIgnoreCase("B")) return bencodeText(text, charset); // BASE64
		else return qencodeText(text, charset); // quoted-printable
	}

	/* ********************************* ENCODE ********************************* */
	/*
	 * The text first character-encoded using the platform's default charset, then transfer-encoded using 
	 * either the Q encoding. The resulting bytes are then returned as a Unicode string containing 
	 * only ASCII characters. 
	 * 
	 * @param text Unicode string 
	 * @return Unicode string containing only US-ASCII characters 
	 */
	private static String qencodeText(String text, String charset) {
		String prefix = "=?" + charset + "?Q?";
		String suffix = "?=";
		int hl = prefix.length() + suffix.length(), wl, length = 0;
		int maxlength = (int)Math.floor((splitLinesAt - hl) / 5.1); // 15% encoded chars (max 4 bytes giving 3 chars each) + 85% ascii chars
		
		StringBuffer buffer = new StringBuffer(prefix);
		boolean first = true;
		for (String word : text.split(" ")) {
			if (!first) { // SPACE => '_' 
				buffer.append("_"); 
				length++;
			} else first = false;
			
			String[] tokens = split(word, maxlength);
			for (int j=0; j<tokens.length; j++) {
				String encodedWord = qencodeWord(tokens[j], charset);
				wl = encodedWord.length();
				if (length + wl + hl >= 75) {
					buffer.append(suffix + SPACE + prefix + encodedWord);
					length = wl;
				} else {
					buffer.append(encodedWord);
					length += wl + 1;
				}
			}
		}
		if (text.endsWith(" "))	buffer.append('_');
		buffer.append(suffix);
		return buffer.toString();
	}
	
	/*
	 * The text first character-encoded using the platform's default charset, then transfer-encoded using 
	 * only ASCII characters. 
	 * 
	 * @param text Unicode string 
	 * @return Unicode string containing only US-ASCII characters 
	 */
	private static String bencodeText(String text, String charset) {
		String prefix = "=?" + charset + "?B?";
		String suffix = "?=";
		int hl = prefix.length() + suffix.length();
		
		boolean first = true; 
		StringBuffer buffer = new StringBuffer();
		for (String line : split(text, (int)Math.floor(splitLinesAt*.75)-hl)) {
			if (!first) buffer.append(" ");
			else first = false;
			buffer.append(prefix);
			
			ByteBuffer bytes = Charset.forName(charset).encode(line);
			
			// add padding bytes if necessary
			int l = bytes.remaining(), paddingCount = (3 - (l % 3)) % 3;
			byte[] b = new byte[l+paddingCount];
			bytes.get(b, 0, l);

			// process 3 bytes at a time, churning out 4 output bytes
			for (int i = 0; i < b.length; i += 3) {
				int j = ((b[i] & 0xff) << 16) + ((b[i+1] & 0xff) << 8) +  (b[i+2] & 0xff);
				buffer.append(base64code[(j >> 18) & 0x3f]);
				buffer.append(base64code[(j >> 12) & 0x3f]);
				buffer.append(base64code[(j >> 6) & 0x3f]);
				buffer.append(base64code[j & 0x3f]);
			}
			// replace encoded padding nulls with "="
			String padding = "===".substring(0, paddingCount);
			buffer.replace(buffer.length() - paddingCount, buffer.length(), padding);
			
			buffer.append(suffix);
		}
		return buffer.toString();
	}

	/*
	 * The word is first character-encoded using the platform's default charset, then transfer-encoded 
	 * using either the Q encoding. The resulting bytes are then returned as a Unicode string 
	 * containing only ASCII characters.
	 *  
	 * @param word Unicode string
	 * @return Array of Unicode strings containing only US-ASCII characters.
	 */
	private static String qencodeWord(String word, String charset) {
		StringBuffer buffer = new StringBuffer();
		ByteBuffer bytes = Charset.forName(charset).encode(word);
		for (int i=0; i<bytes.remaining(); i++) {
			byte b = bytes.get(i);
			char c = (char)b;
			if (c == '=') buffer.append("=3D");
			else if (c == '?') buffer.append("=3F");
			else if (c == '_') buffer.append("=5F");
			else if (c == SPACE) buffer.append("=20");
			else if (c == HTAB) buffer.append("=09");
			else if (c < 128) buffer.append(c);
			else buffer.append("=" + toHexString(b));				
 		}
		return buffer.toString();
	}
	
	/*
	 * encoded-word = "=?" charset "?" encoding "?" encoded-text "?="
	 */
	private static boolean isEncodedWord(String word) {
		boolean isEncodedWord = false;
		if (word.length() > 6 && word.startsWith("=?") && word.endsWith("?=")) {
			int idx1 = word.indexOf("?", 2);
			if (word.length() >= idx1+4) {
				int idx2 = word.indexOf("?", idx1+1);
				isEncodedWord = idx2 > idx1 && idx2 < word.length()-2;
			}
		}
		return isEncodedWord;
	}

	/* ********************************* DECODE ********************************* */
	/*
	 * The string is parsed using the rules in RFC 2047 and RFC 2231 for parsing an "encoded-word". 
	 * If the parse fails, a ParseException is thrown. Otherwise, it is transfer-decoded, and then 
	 * charset-converted into Unicode. If the charset-conversion fails, an UnsupportedEncodingException is thrown.
	 * 
	 * @param eword the encoded value: "=?" charset "?" encoding "?" encoded-text "?="
	 * @return
	 * @throws ParseException if the string is not an encoded-word as per RFC 2047 and RFC 2231. 
	 * @throws UnsupportedEncodingException if the charset conversion failed.
	 */
	private static String decodeWord(String eword) throws ParseException, UnsupportedEncodingException {
		int idx1 = eword.indexOf("?", 2), idx2 = eword.indexOf("?", idx1+1);
		String charsetname = eword.substring(2, idx1);
		String text = eword.substring(idx2+1, eword.length()-2);
		String encoding = eword.substring(idx1+1, idx2);
		
		Charset charset = null;
		try {
			charset = Charset.forName(charsetname);
		} catch (Exception e) {
			throw new UnsupportedEncodingException(e.getClass().getName() + ": "+charsetname);
		}
		
		byte[] buffer = null;
		if (encoding.equalsIgnoreCase("b")) 
			buffer = bdecodeWord(text);
		else if (encoding.equalsIgnoreCase("q")) 
			buffer = qdecodeWord(text);
		else
			throw new UnsupportedEncodingException("Unsupported encoding: "+encoding);
		
		return charset.decode(ByteBuffer.wrap(buffer)).toString();
	}
	
	private static byte[] bdecodeWord(String eword) {
		int i = eword.length()-1, paddingCount = 0;
		while (eword.charAt(i) == '=') {
			paddingCount++; i--;
		}
		
		byte[] buffer = new byte[(int)eword.length()*3/4];
		int idx = 0;
		for (i=0; i<eword.length(); i += 4) {
			int j = (indexOf(eword, i, paddingCount) << 18) + (indexOf(eword, i+1, paddingCount) << 12) + 
			(indexOf(eword, i+2, paddingCount) << 6) + indexOf(eword, i+3, paddingCount);
			buffer[idx++] = (byte)(j >> 16);
			buffer[idx++] = (byte)(j >> 8);
			buffer[idx++] = (byte)j;
		}
		
		// remove padding
		byte[] b = new byte[buffer.length-paddingCount];
		System.arraycopy(buffer, 0, b, 0, b.length);
		return b;
	}
	
	private static int indexOf(String s, int i, int paddingCount) {
		if (i < s.length() - paddingCount)
			return indexOf(s.charAt(i), base64code);
		else
			return 0;
	}

	private static byte[] qdecodeWord(String eword) throws ParseException {
		byte[] buffer = new byte[eword.length()];
		int j = 0;
		for (int i=0; i<eword.length(); i++) {
			char c = eword.charAt(i);
			if (c == '=') {
				if (i+2 >= eword.length())
					throw new ParseException("Invalid character '='", i);
				buffer[j++] = parseHexString(eword.substring(i+1, i+3));
				i += 2;
			} else if (c == '_') {
				buffer[j++] = (byte)SPACE;
			} else
				buffer[j++] = (byte)c;
		}
		
		byte[] b = new byte[j];
		System.arraycopy(buffer, 0, b, 0, j);
		return b;
	}
	
	/* ********************************* UTILS ********************************* */
	/*
	 * split a word into short tokens 
	 */
	private static String[] split(String word, int length) {
		if (word.length() <= length)
			return new String[]{ word };
		int idx = 0, n = word.length() / length;
		if (word.length() % length != 0) n++;
		String[] tokens = new String[n];
		for (int i=0; i<n-1; i++) {
			tokens[i] = word.substring(idx, idx + length);
			idx += length;
		}
		tokens[n-1] = word.substring(idx);
		return tokens;
	}
	
	/*
	 * split text into 'linear-white-space'-separated words
	 */
	private static String[] split(String text) {
		List<String> list = new ArrayList<String>();
		int beginIndex = 0;
		int endIndex = nextLinearWhiteSpace(text, 0);
		while (endIndex >= 0) {
			list.add(text.substring(beginIndex, endIndex));
			beginIndex = nextNonLinearWhiteSpace(text, endIndex);
			endIndex = nextLinearWhiteSpace(text, beginIndex);
		}
		if (beginIndex >= 0)
			list.add(text.substring(beginIndex));
		String[] words = new String[list.size()];
		list.toArray(words);
		return words;
	}
	
	/*
	 * Gives the index of the next linear-white-space, starting at beginIndex
	 */
	private static int nextLinearWhiteSpace(String text, int beginIndex) {
		int idx = -1, i = beginIndex; 
		if (i >= 0) 
			while (i < text.length() && idx < 0) {
				if (isLinearWhiteSpace(text, i)) 
					idx = i;
				i++;
			}
		return idx;
	}
	
	/*
	 * Gives the index of the next non linear-white-space, starting at beginIndex
	 */
	private static int nextNonLinearWhiteSpace(String text, int beginIndex) {
		int idx = -1, i = beginIndex;
		if (i >= 0) 
			while (i < text.length() && idx < 0) {
				if (!isLinearWhiteSpace(text, i)) 
					idx = i;
				i++;
			}
		return idx;
	}
	
	/*
	 * linear-white-space ::=  1*([CRLF] LWSP-char)
	 */
	private static boolean isLinearWhiteSpace(String text, int index) {
		return isLWSP(text, index) || (isCRLF(text, index) && isLWSP(text, index+2));
	}
	
	/*
	 * CRLF               ::=  CR LF
	 */
	private static boolean isCRLF(String text, int index) {
		if (index >= 0 && index+1 < text.length()) 
			return text.charAt(index) == CR && text.charAt(index+1) == LF;
		return false;
	}
	
	/*
	 * LWSP-char          ::=  SPACE | HTAB
	 */
	private static boolean isLWSP(String text, int index) {
		if (index >= 0 && index < text.length()) 
			return text.charAt(index) == SPACE || text.charAt(index) == HTAB;
		return false;
	}
	
	private static byte parseHexString(String s) {
		int low = indexOf(s.charAt(1), hexdigits);
		int high = indexOf(s.charAt(0), hexdigits);
		return (byte)((high << 4) + low);
	}

	private static String toHexString(byte b) {
		int low  = b & 0x0F;
		int high = (b & 0xF0) >> 4;
		return String.valueOf(hexdigits[high]) + String.valueOf(hexdigits[low]);
	}
	
	private static int indexOf(char c, char[] digits) {
		for (int i=0; i<digits.length; i++)
			if (c == digits[i])
				return i;
		return -1;
	}
}
