/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.util.Properties;

/**
 * <p>The JVM that hosts DoDWAN needs to know minimal information to function properly, 
 * namely the name of the DoDWAN host and the name of some working directories. 
 * This class provides read-only variables for this, whose values are coming from the system
 * environment. Some of these variables are normally set from values supposed to be passed 
 * to the JVM via Java properties.
 * 
 * <p>The name of the DoDWAN host is usually the actual name of the host running DoDWAN but you 
 * may use any another name (for simulation purposes for example), provided that it is unique 
 * among the participants in the network. If no DoDWAN host is specified, it is considered that 
 * the DoWDAN host is the he actual name of the host running DoDWAN (returned by the gethostname Java method).

 * <p>DoDWAN uses three working directories:
 * <ul>
 *  <li>The DoDWAN home directory: this is the directory where are placed the JAR file containing 
 *      the code of DoDWAN (in subdirectory lib) and the default configuration file 
 *      (in subdirectory conf). </li>
 *  <li> The DoDWAN var directory: this is the directory in which DoDWAN stores its cache 
 *       and other internal information. </li>
 *  <li> The DoDWAN base directory: this is the directory in which the user data is stored 
 *       (eg the user preferences for DoDWAN configuration). </li>
 *  </ul>
 *
 * <p>DoDWAN var and the DoDWAN base directories may not exist when DoDWAN starts. In this case, 
 * they will be automatically created. Files are not stored directly under the DoDWAN var 
 * and the DoDWAN base directories. They are actually placed in a subdirectory that has the name
 * of the DoDWAN host, so that multiple instances of DoDWAN can run concurrently on the same 
 * machine for simulation purposes. 
 * 
 * <p> It is the responsibility of the deployer to ensure that the DoDWAN host as well as the 
 * DoDWAN home, DoDWAN var and DoDWAN base directories are designated when launching the JVM, 
 * through the dodwan.host, dodwan.home, dodwan.var and dodwan.base properties (typically by 
 * using the -D option). 
 * 
 * <p>The definition of three distinct directories for DoDWAN home, DoDWAN var and DoDWAN base is 
 * expected in order to allow attributing distinct access permissions for each of these directories, 
 * which may be useful in multi-user systems. In single-user systems, a simpler alternative is
 * possible via the setting of only one directory, the Root directory, instead of the three DoDWAN home, 
 * DoDWAN var and DoDWAN base ones. In this case DoDWAN home, DoDWAN var and DoDWAN base are 
 * automatically set as subdirectories of the provided Root directory. 
 * The Root directory is supposed to be passed to the JVM through the dodwan.root property. 
 * and it is created if it does no exists.
 * 
 * <p>If neither the Root directory nor the triplet (Home, Var, Base directories) is specified, it is 
 * considered that the Root directory is the "DoDWAN" subdirectory of the homedir (as found 
 * in the user.home system property).
 * 
 * <p>In some systems, the launching of the JVM cannot be parameterized so none of the properties 
 * dodwan.host dodwan.root, dodwan.base, dodwan var, and dodwan.base can be passed to the JVM. 
 * As a workaround, method setPathsAsJvmEnv() can be used. Be aware that this method must be 
 * called before any other call to another method of DoDWAN middleware.
 */

public class SystemEnvironment {
	
	/**
	 * Version number of DoDWAN
	 */
	public static String version; 
	
    
    /**
     * Root of the directory tree where the DoDWAN package has been installed.
     */
    public static String rootPath = System.getProperty("dodwan.root");
    	
    /**
     * Directory where the all the files coming from the DoDWAN package can be found.
     */
    public static String homePath = System.getProperty("dodwan.home");

    /**
     * Directory in which DoDWAN stores its cache and other non-user-specific internal information.
     */
    public static String varPath = System.getProperty("dodwan.var");
 
    /**
     * Directory in which the user data is stored (eg the user preferences for DoDWAN configuration).
     */
    public static String basePath = System.getProperty("dodwan.base");

    /**
     * Name of the DoDWAN host.
     */
    public static String host = System.getProperty("dodwan.host");
	

    static {
		version = "unknown";
    	Properties  props = new Properties();
		InputStream in = SystemEnvironment.class.getResourceAsStream("/META-INF/maven/casa.dodwan/dodwan/pom.properties");
    	try {
			if (in != null) 
				props.load(in);
    	} catch (IOException e) {
    		System.err.println("No POM properties file found in jar");
    	}
    	version = props.getProperty("version");



    	if ((homePath == null) || (varPath == null) || (basePath == null)) {
    		if (rootPath != null) {
    			homePath = rootPath + File.separator + "home"; 
    			System.setProperty("dodwan.home", rootPath + File.separator + "home");

    			varPath = rootPath + File.separator + "var";
    			System.setProperty("dodwan.var", rootPath + File.separator + "var");

    			basePath = rootPath + File.separator + "base";
    			System.setProperty("dodwan.base", rootPath + File.separator + "base");
    		}
    		else {
    			String homeRoot = System.getProperty("user.home")  + File.separator + 
    					"DoDWAN" + File.separator;
    			homePath = homeRoot + "home";
    			varPath =  homeRoot + "var";
    			basePath = homeRoot + "base";
    		}
    	}	

    	if (host == null) {
    		try {
    			InetAddress localHost = InetAddress.getLocalHost();
    			host = localHost.getHostName();
    		} 
    		catch (Exception e) {
    			System.err.println("No host name found");
    		}

    	}
    }



    /**
     * @undocumented
     */
    public static String varHostPath  = varPath  + File.separator + "host" + File.separator + host ;

    /**
     * @undocumented
     */
    public static String baseHostPath = basePath + File.separator + "host" + File.separator + host ;
    
    
 
    
    /**
     * Defines the host variable. This method should be called only in case the JVM process environment 
     * does not provide the right one, and before any method of other classes of the DoDWAN middleware. 
     */
    public static void setHostAsJvmEnv(String h) {
        host = h;
   	    System.setProperty("dodwan.host", host);
   }

    
    /**
     * Defines the Root path. This method should be called only in case the JVM process environment 
     * does not provide the right one, and before any method of other classes of the DoDWAN middleware. 
     * Other paths are built from this Root path.
     * @param rp the Root path
     */
    public static void setPathsAsJvmEnv(String rp) {
    	
    	rootPath = rp;
	   
	    homePath = rootPath + File.separator + "home"; 
		System.setProperty("dodwan.home", rootPath + File.separator + "home");
		
		varPath = rootPath + File.separator + "var";
		System.setProperty("dodwan.var", rootPath + File.separator + "var");

		basePath = rootPath + File.separator + "base";
		System.setProperty("dodwan.base", rootPath + File.separator + "base");

		varHostPath  = varPath  + File.separator + "host" + File.separator + host ;	    
        baseHostPath = basePath + File.separator + "host" + File.separator + host ;
    }
   

	    
    /**
     * Defines the Home, Var and Base paths. This method should be called only in case the JVM process environment 
     * does not provide the right one, and before any method of other classes of the DoDWAN middleware.
     * @param hp the Home path
     * @param vp the Var path
     * @param bp the Base path
     */
    public static void setPathsAsJvmEnv(String hp, String vp, String bp) {
	    homePath = hp; 
		System.setProperty("dodwan.home", rootPath + File.separator + "home");
		
		varPath = vp;
		System.setProperty("dodwan.var", rootPath + File.separator + "var");

		basePath = bp;
		System.setProperty("dodwan.base", rootPath + File.separator + "base");

		varHostPath  = varPath  + File.separator + "host" + File.separator + host ;	    
        baseHostPath = basePath + File.separator + "host" + File.separator + host ;
    }
 
   
    
    /**
     * 
     * @undocumented
     *
     * Returns a path as a File. The complete path (including parent directories)
     * are created if it does not exist yet.
     * @param spath The name of the path
     * @return the path or null if a problem occurred while creating the path
     */
    public static File getPath(String spath) {
    	File p = new File(spath);
    	try {
    		if (! p.exists()) {
    			p.mkdirs();
    		}
    	}
    	catch (SecurityException se) {
    		p = null;
    	}
    	return p;
    }
  
    
    /**
     * 
     * @undocumented
     *
     * Returns the Dodwan Root path (the path is created if it not already done) 
     * @return the Dodwan Root path (or null if a problem occurred while creating the path)
     */
    public static File getRootPath() {
    	return getPath(rootPath);
    }

    /**
     * 
     * @undocumented
     *
     * Returns the Dodwan Home path (the path is created if it not already done) 
     * @return the Dodwan Home path (or null if a problem occurred while creating the path)
     */
    public static File getHomePath() {
    	return getPath(homePath);
    }
    
    /**
     * 
     * @undocumented
     *
     * Returns the Dodwan Var path (the path is created if it is not already done) 
     * @return the Dodwan Var path (or null if a problem occurred while creating the path)
     */
    public static File getVarPath() {
    	return getPath(varPath);
    }
  	
    /**
     * 
     * @undocumented
     *
     * Returns the Dodwan Base path (the path is created if it is not already done) 
     * @return the Dodwan base path (or null if a problem occurred while creating the path)
     */
    public static File getBasePath() {
    	return getPath(basePath);
    }
  
    /**
     * 
     * @undocumented
     *
     * Returns the Dodwan VarHost path (the path is created if it is not already done) 
     * @return the Dodwan VarHost path (or null if a problem occurred while creating the path)
     */
    public static File getVarHostPath() {
    	return getPath(varHostPath);
    }
 
    /**
     * 
     * @undocumented
     *
     * Returns the Dodwan BaseHost path (the path is created if it is not already done) 
     * @return the Dodwan BaseHost path (or null if a problem occurred while creating the path)
     */
    public static File getBaseHostPath() {
    	return getPath(baseHostPath);
    }
     
}
